#include "main.h"
#include "read_maf.h"
#include "util.h"
#include "util_input.h"
#include "util_gen.h"
#include "analysis.h"

int debug_mode;
char S[BIG], T[BIG];
char S1[BIG], T1[BIG];

int main(int argc, char **argv)
{
	struct DotList *content_ortho;
	struct DotList *position_ortho;
	int *num_content, *num_position;
	int *size1, *size2;
	int count = 0;
	char species[100], species2[100];
	FILE *f, *g;
	struct exons_list *exons, *genes; 
	int *num_exons, *num_genes;
	float content_avg_pid = (float)0; 
	float position_avg_pid = (float)0; 

	debug_mode = FALSE;
	if( argc == 4 ) {
		debug_mode = TRUE;
	}
	else if( argc != 3 ) {
		fatal("args: content_ortho_maf position_ortho_maf annot.codex\n");
	}

	size1 = (int *) ckalloc(sizeof(int));
	size2 = (int *) ckalloc(sizeof(int));
	num_content = (int *) ckalloc(sizeof(int));
	num_position = (int *) ckalloc(sizeof(int));
	*size1 = 0;
	*size2 = 0;
	*num_content = 0;
	*num_position = 0;
	count = count_local_algns(argv[1], species, species2);
	if( count > 0 ) {
		content_ortho = (struct DotList *) ckalloc(count * sizeof(struct DotList));
		initialize_algns(content_ortho, 0, count);

	 	read_maf(argv[1], O_MODE, content_ortho, num_content, size1, size2);
	}

	if( count != (*num_content) ) {
		fatalf("counting error 1: %d vs. %d\n", count, *num_content);
	}

	count = count_local_algns(argv[2], species, species2);
	if( count > 0 ) {
		position_ortho = (struct DotList *) ckalloc(count * sizeof(struct DotList));
		initialize_algns(position_ortho, 0, count);

	 	read_maf(argv[2], O_MODE, position_ortho, num_position, size1, size2);
	}

	if( count != (*num_position) ) {
		fatalf("counting error 2: %d vs. %d\n", count, *num_position);
	}

	num_exons = (int *) ckalloc(sizeof(int));
	*num_exons = 0;
	num_genes = (int *) ckalloc(sizeof(int));
	*num_genes = 0;

	count = count_genes(argv[3], species, species2);
	genes = (struct exons_list *) ckalloc(count * sizeof(struct exons_list));

	count = count_exons(argv[3], species, species2);
	exons = (struct exons_list *) ckalloc(count * sizeof(struct exons_list));
	read_only_exons(exons, num_exons, genes, num_genes, argv[3], species, species2);

	if( count != (*num_exons) ) {
		fatalf("counting error 3: %d vs. %d\n", count, *num_exons);
	}

	if( (*num_content) > 0 ) {
		content_avg_pid = cal_avg_pid(content_ortho, *num_content);
	}

	if( (*num_position) > 0 ) {
		position_avg_pid = cal_avg_pid(position_ortho, *num_position);
	}
	
	f = ckopen(argv[1], "r");
	g = ckopen(argv[2], "r");

//	get_numbers(content_ortho, *num_content, position_ortho, *num_position, exons, *num_exons, f, g, *size1, *size2); 
	get_numbers(content_ortho, *num_content, position_ortho, *num_position, f, g, *size1, *size2); 

	fclose(g);
	fclose(f);
	free(size1);
	free(size2);

	if( (*num_content) > 0 ) {
		free(content_ortho);
	}

	if( (*num_position) > 0 ) {
		free(position_ortho);
	}

	free(num_content);
	free(num_position);
	free(exons);
	free(genes);
	free(num_exons);
	free(num_genes);
	return(EXIT_SUCCESS);
}

